import logging
import functools

from typing import Union, Callable, Optional

import paramiko
from typeguard import typechecked

from . import _ssh_client

logger = logging.getLogger()


@typechecked
def initialize(
    *,
    address: str,
    base_prompt: str,
    port: Union[int, str] = 22,
    username: Optional[str] = None,
    password: Optional[str] = None,
    private_key: Optional[str] = None,
    passphrase: Optional[str] = None,
    timeout: Optional[int] = 10,
    shell_timeout: Optional[int] = 5,
    verify_host_key: bool = False,
) -> Callable:
    """ Initialize an SSH session and get a connection context manager.

    :param address: The hostname or IP address of the device to connect to.
    :param base_prompt: A regular expression matching the command prompt on the device.
        This is used to detect when a command has finished executing, and to accurately
        extract output from a command.
    :param port: The listening SSH port (usually 22).
    :param username: Optional username for :ref:`index:Authentication`.
    :param password: Optional password for :ref:`index:Authentication`. If supplied, can be used for
        private key decryption if a key is supplied with no passphrase argument.
    :param private_key: Optional private key to use for key :ref:`index:Authentication`.
    :param passphrase: Used for decrypting the private key (see :ref:`index:Authentication`).
    :param timeout: The timeout to use when opening SSH connections.
    :param shell_timeout: The timeout to wait for the SSH shell buffer.
    :param verify_host_key: Whether to verify the device host key or not when connecting.

    :raises exceptions.AuthenticationFailed: if authentication using the supplied credentials failed.
    :raises exceptions.ConnectionFailed: if connectivity to the supplied address failed.
    :raises exceptions.PrivateKeyError: if a supplied private key fails to load.

    :returns: A context manager that, when called using the `with` statement, yields a connector
        for running SSH commands.

    The returned context manager should only be called using a `with statement
    <https://docs.python.org/3.8/reference/compound_stmts.html#with>`_.
    All SSH operations must happen inside the `with` block. You do not need to know how to
    open or close an SSH connection - the context manager handles all of this for you.

    You can re-use an initialized session as often as necessary. This makes it easier to
    separate offline programming logic from SSH operations as much as possible.

    See below for a basic snippet, or :ref:`index:Authentication` and :ref:`index:Example Scripts` for
    more detailed examples.

        .. code-block:: python

            get_connector = client.initialize(
                address="10.9.10.134",
                username="user123",
                password="mUh5f&uid@",
                base_prompt=r".*@\\w+.*\\$\\s{1}",
            )
            with get_connector() as connector:
                data = client.run(connector, command='ls -lsah')

            # print the command output, and the string matched by the base_prompt.
            print(data.output)
            print(data.match_info.string)
    """
    return functools.partial(
        _ssh_client._connect,
        address=address,
        port=port,
        username=username,
        password=password,
        private_key=private_key,
        passphrase=passphrase,
        timeout=timeout,
        base_prompt=base_prompt,
        shell_timeout=shell_timeout,
        missing_host_key_policy=paramiko.RejectPolicy() if verify_host_key else paramiko.WarningPolicy(),
    )


@typechecked
def run(
    connection: Callable,
    *,
    command: str,
    prompt: Optional[Union[str, bytes]] = None,
    use_regex: Optional[bool] = True,
    output_raw: Optional[bool] = False,
) -> _ssh_client.CommandReturn:
    """
    Execute a command on the target device, and return the output and prompt match
    information as a :ref:`return_objects:Command Return Tuple`.

    :param connection: An SSH command runner.
    :param command: The command to run.
    :param prompt: Optional. An alternate prompt to look for after running
        the command. Useful if the command triggers a confirmation or password
        prompt. This prompt is processed as a regular expression by default,
        but simple string searching can be used by supplying `use_regex=False`.
    :param use_regex: Optional, defaults to True. Denotes whether the alternate
        prompt, if supplied, should be processed as a regular expression. If this
        argument is `False`, the prompt will be processed with a simple string
        search. If this argument is provided with no prompt supplied, it will
        be ignored.
    :param output_raw: Optional. Defaults to False. Denotes whether to clean the
        response before returning it. When `False`, the echoed SSH command will be
        stripped from the response, and the response will be decoded to utf-8. When
        `True`, the command will not be stripped and the return type will be bytes.
        This is mostly useful for debugging.

    :raises exceptions.AuthenticationFailed: if authentication using the supplied credentials failed.
    :raises exceptions.ConnectionFailed: if connectivity to the supplied address failed.
    :raises exceptions.OperationTimeout: if the SSH shell timed out while attempting to run the command.
    :raises exceptions.PatternNotFound: if the SSH client fails to find the shell prompt after executing the command.
    :raises exceptions.InvalidRegex: if a supplied prompt fails to compile to a valid regular expression.

    :returns: A :ref:`return_objects:Command Return Tuple`.

    Running a basic command, and printing the output and matched prompt:

        .. code-block:: python

            command_return = client.run(connection, command='ls -lsah')
            print(command_return.output)
            print(command_return.match_info.string)

    Running a command that triggers a password prompt, and supplying the password:

        .. code-block:: python

            client.run(
                connection,
                command="enable",
                prompt="Password:",
                use_regex=False,
                command='enable'
            )
            client.run(connection, command="dfh843hf*(h")

    Running a command and saving the raw output:

        .. code-block:: python

            raw_output = client.run(connection, command='ls -lsah', output_raw=True).output
    """
    prompt_fn = (
        {True: _ssh_client._regex_search, False: _ssh_client._string_search}.get(use_regex)(prompt) if prompt else None
    )
    output, match_data = connection(command=command, alt_prompt_fn=prompt_fn)
    return _ssh_client._command_return(  # type: ignore
        output=_ssh_client._clean_response(response=output, command=command) if not output_raw else output,
        match_info=match_data,
    )
