import functools
import stat

from datetime import datetime, timezone
from typing import Callable, List


from ppa import ui
from ssh import sftp_client, exceptions as ssh_exceptions

from .records import DirectoryItem, SFTPServer


def sftp_connection(func: Callable) -> Callable:
    @functools.wraps(func)
    def _connection_wrapper(sftp_server: SFTPServer, *args, **kwargs) -> Callable:
        try:
            with sftp_client.initialize(**sftp_server)() as connection:
                return func(connection, *args, **kwargs)
        except (ssh_exceptions.AuthenticationFailed, ssh_exceptions.ConnectionFailed) as e:
            ui.output_error(str(e))
            raise e
        except Exception as e:
            ui.output_error("An unhandled connection error occurred. Please view the logs for more information.")
            raise e

    return _connection_wrapper


def create_directory_items(func: Callable) -> Callable:
    @functools.wraps(func)
    def _execute(*args, **kwargs) -> List[DirectoryItem]:
        return [
            DirectoryItem(
                name=f.filename,
                is_dir=stat.S_ISDIR(f.st_mode),
                size=f.st_size,
                last_modified=datetime.fromtimestamp(f.st_mtime, tz=timezone.utc).strftime("%H:%M %d/%m/%Y"),
            )
            for f in func(*args, **kwargs)
        ]

    return _execute
