r"""
Client

This module contains actions for listing directories on & uploading/downloading files to/from an SFTP server.
"""
import os
import re
import logging
from typing import List, Optional, Union

from ppa import ui
from ssh import sftp_client

from ._client import sftp_connection, create_directory_items
from .records import DirectoryItem, SFTPServer


TABLE_HEADER = ["Name", "Is Directory", "Size (bytes)", "Last Modified"]
TABLE_FIELDS = ["name", "is_dir", "size", "last_modified"]
SHARED_WORKSPACE = "/workspace/share"


logger = logging.getLogger(__name__)


class InvalidExpression(Exception):
    pass


def _safe_table_get(key, dictionary):
    return str(dictionary.get(key)) if dictionary.get(key) is not None else ""


def _change_working_directory(server, directory):
    if directory:
        server.chdir(directory)


@sftp_connection
@create_directory_items
def list_directory(sftp_server: SFTPServer, path: str = ".", working_directory: Optional[str] = None) -> List[DirectoryItem]:
    """List items in a directory

??? info "Input"

    - __sftp_server:__ an [SFTPServer](../inputs_outputs/#sftpserver) dictionary

    - __path:__ optional path to list items from (defaults to current working directory `.`)

    - __working_directory:__ optional remote directory from which to run the command (defaults to the login directory)

??? success "Output"

    A list of [DirectoryItem](../inputs_outputs/#directoryitem) dictionaries.

??? example

    ```YAML linenums="1"
    - sftp.client.list_directory:
      load:
        sftp_server: sftp_server
      save: directory_listing
    ```

"""
    logger.info(f"listing directory path '{path}'")
    _change_working_directory(sftp_server, working_directory)
    return sftp_server.listdir_attr(path)


@sftp_connection
@create_directory_items
def search_directory(
    sftp_server: SFTPServer, expression: str, path: str = ".", working_directory: Optional[str] = None
) -> List[DirectoryItem]:
    r"""Search a directory for items using a regular expression.

??? info "Input"

    - __sftp_server:__ an [SFTPServer](../inputs_outputs/#sftpserver) dictionary

    - __expression:__ regular expression to use in the search

    - __path:__ optional path to list files in (defaults to current working directory `.`)

    - __working_directory:__ optional remote directory from which to run the command (defaults to the login directory)

??? success "Output"

    A list of [DirectoryItem](../inputs_outputs/#directoryitem) dictionaries.

??? example

    ```YAML linenums="1"
    - sftp.client.search_directory:
        expression: .*\.json
      load:
        sftp_server: sftp_server
      save: json_files
    ```

    """
    try:
        regex = re.compile(expression)
    except re.error as e:
        message = f"The supplied regular expression '{expression}' is invalid: {e.args[0]}."
        ui.output_error(message)
        raise InvalidExpression(message)

    logger.info(f"searching for items matching pattern '{expression}' in path '{path}'")
    _change_working_directory(sftp_server, working_directory)
    return list(filter(lambda item: re.search(regex, item.filename), sftp_server.listdir_attr(path)))


@sftp_connection
def delete_file(sftp_server: SFTPServer, name: str, working_directory: Optional[str] = None):
    """Delete a file on an SFTP server.

??? info "Input"

    - __sftp_server:__ an [SFTPServer](../inputs_outputs/#sftpserver) dictionary

    - __name:__ the name of the directory item to delete

    - __working_directory:__ optional remote directory from which to run the command (defaults to the login directory)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - sftp.client.delete_file:
        name: example_file.json
      load:
        sftp_server: sftp_server
    ```
    """
    logger.info(f"deleting file '{name}'")
    _change_working_directory(sftp_server, working_directory)
    sftp_server.remove(name)


@sftp_connection
def delete_files(sftp_server: SFTPServer, names: List[str], working_directory: Optional[str] = None):
    """Delete multiple files on an SFTP server.

??? info "Input"

    - __sftp_server:__ an [SFTPServer](../inputs_outputs/#sftpserver) dictionary

    - __names:__ a list containing the name of each directory item to delete

    - __working_directory:__ optional remote directory from which to run the command (defaults to the login directory)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - sftp.client.delete_files:
        names:
          - json_file.json
          - text_file.txt
      load:
        sftp_server: sftp_server
    ```
    """
    logger.info(f"deleting files '{names}'")
    _change_working_directory(sftp_server, working_directory)
    return list(map(lambda name: sftp_server.remove(name), names))


@sftp_connection
def upload_file(sftp_server: SFTPServer, source: str, destination: str):
    """Upload a file to the SFTP server.

??? info "Input"

    - __sftp_server:__ an [SFTPServer](../inputs_outputs/#sftpserver) dictionary

    - __source:__ the name of the local file to upload

    - __destination:__ the destination path on the SFTP server (_including the file name_)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - sftp.client.upload_file:
        source: playbook_file.txt
        destination: playbook_file.txt
      load:
        sftp_server: sftp_server
    ```
    """
    logger.info(f"uploading file from '{source}' to '{destination}'")
    return sftp_client.upload_file(
        sftp_server, source=os.path.join(SHARED_WORKSPACE, source.split("/")[-1]), destination=destination
    )


@sftp_connection
def download_file(sftp_server: SFTPServer, source: str, destination: str):
    """Download a file from an SFTP server.

??? info "Input"

    - __sftp_server:__ an [SFTPServer](../inputs_outputs/#sftpserver) dictionary

    - __source:__ the remote path of the file to download

    - __destination:__ the name of the local file to create

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - sftp.client.download_file:
        source: remote_directory/remote_file.txt
        destination: local_file.txt
      load:
        sftp_server: sftp_server
    ```
    """
    logger.info(f"downloading file from '{source}' to '{destination}'")
    return sftp_client.download_file(
        sftp_server, source=source, destination=os.path.join(SHARED_WORKSPACE, destination.split("/")[-1])
    )


def output_directory_table(text: str, directory_items: Union[List[DirectoryItem], DirectoryItem]):
    r"""Display directory items in a table.

The table will have the following columns:

- Name
- Is Directory
- Size (bytes)
- Last Modified

??? info "Input"

    - __directory_items:__ a single or list of [DirectoryItem](../inputs_outputs/#directoryitem) dictionaries to display in a table

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - sftp.client.search_directory:
        expression: .*\.json
      load:
        sftp_server: sftp_server
      save: json_files

    - sftp.client.output_directory_table:
        text: JSON Files
      load:
        directory_items: json_files
    ```

    """
    directory_items = [directory_items] if not isinstance(directory_items, list) else directory_items
    return ui.output_table(
        text=text,
        header=TABLE_HEADER,
        data=[[_safe_table_get(field, item) for field in TABLE_FIELDS] for item in directory_items],
    )
